# Simple comparison of SDKN approach vs NN on a 1D toy problem.

# Some imports and random seed
import math
import time

import numpy as np
import torch
import torch.optim as optim
from matplotlib import pyplot as plt

from utils.models import NN, SDKN

torch.random.manual_seed(2)

# Create SDKN model with some arbitrary centers
centers = torch.tensor([[-1, -.5, 0, .5, 1]]).t()
model_SDKN = SDKN(centers=centers, dim_input=1, dim_output=1)
model_NN = NN(dim_input=1, dim_output=1)

# Print total number of parameters
total_params1 = sum(p.numel() for p in model_SDKN.parameters() if p.requires_grad)
total_params2 = sum(p.numel() for p in model_NN.parameters() if p.requires_grad)
print('nr params model1 = {}.'.format(total_params1))
print('nr params model2 = {}.'.format(total_params2))

# Create training and test data
X_train = torch.linspace(-1, 1, 1000).view(-1, 1)
X_test = torch.linspace(-2, 2, 1000).view(-1, 1)


# Define some target function for approximation
def f_func(x):
    return torch.sin(math.pi * x ** 2)


y_train = f_func(X_train)

# Set up training details
optimizer1 = optim.Adam(model_SDKN.parameters(), lr=.05)
optimizer2 = optim.Adam(model_NN.parameters(), lr=.05)
loss = torch.nn.MSELoss(reduction='mean')

batch_size = 200
n_epoch = int(1e3)
N = X_train.shape[0]
n_batches = int(np.floor(N / batch_size))  # some samples might be lost

lambda_lr_decay = lambda epoch: 1 / (1 + .02 * epoch ** 1.2)
scheduler1 = torch.optim.lr_scheduler.LambdaLR(optimizer1, lr_lambda=lambda_lr_decay)
scheduler2 = torch.optim.lr_scheduler.LambdaLR(optimizer2, lr_lambda=lambda_lr_decay)

# Start the training and keep track of the (training) losses
list_loss1 = []
list_loss2 = []
t1 = time.time()
for idx_epoch in range(n_epoch):  # loop over the dataset multiple times

    shuffle = np.random.permutation(X_train.shape[0])  # shuffle the data set

    for idx_batch in range(n_batches):
        # Set gradients to zero
        optimizer1.zero_grad()
        optimizer2.zero_grad()

        # Select a mini-batch of data
        ind = shuffle[idx_batch * batch_size: min((idx_batch + 1) * batch_size, N)]
        X_batch = X_train[ind, :]

        # forward + backward + optimize
        outputs1 = model_SDKN(X_batch)
        outputs2 = model_NN(X_batch)
        loss1 = loss(outputs1, y_train[ind, :])
        loss2 = loss(outputs2, y_train[ind, :])
        loss1.backward()
        loss2.backward()
        optimizer1.step()
        optimizer2.step()

    t2 = time.time()

    # modify learning rate
    scheduler1.step()
    scheduler2.step()

    timediff = t2 - t1

    # Calculate and store loss on whole training set
    loss_all1 = loss(model_SDKN(X_train), y_train)
    loss_all2 = loss(model_NN(X_train), y_train)
    list_loss1.append(loss_all1.item())
    list_loss2.append(loss_all2.item())

    # Print some training information from time to time
    if idx_epoch % 10 == 9:
        print('{:d}, time: {:.1f}: loss1: {:.3e}, loss2: {:.3e}.'.format(
            idx_epoch + 1, timediff, loss_all1, loss_all2))

# Calculate overall loss after training
loss_overall_model1 = loss(model_SDKN(X_train), y_train)
loss_overall_model2 = loss(model_NN(X_train), y_train)

# Several plots, e.g of prediction, residual, loss decay
plt.figure(1)
plt.clf()
plt.plot(X_test, f_func(X_test), 'k')
plt.plot(X_test, model_SDKN(X_test).detach())
plt.plot(X_test, model_NN(X_test).detach())
plt.legend(['function', 'prediction1', 'prediction2'])
plt.draw()

plt.figure(2)
plt.clf()
plt.plot(X_test, model_SDKN(X_test).detach() - f_func(X_test))
plt.plot(X_test, model_NN(X_test).detach() - f_func(X_test))
plt.legend(['residual1', 'residual2'])
plt.draw()

plt.figure(3)
plt.clf()
plt.plot(list_loss1)
plt.plot(list_loss2)
plt.legend(['loss1', 'loss2'])
plt.xscale('log')
plt.yscale('log')
plt.draw()

plt.show()
