# Implementation of both SDKN and NN in the pytorch-lightning framework.

import pytorch_lightning as pl
import torch
from torch import nn
from utils.utilities import ActivFunc
from utils import kernels
from torch.nn import functional as F
from utils import settings


class Network(pl.LightningModule):
    '''
    General class with training_step, validation_step and test_step method.
    '''

    def __init__(self):
        super().__init__()


    def training_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self(x)
        loss = F.mse_loss(y_pred, y)

        # Some logs
        self.log('train/loss', loss, prog_bar=True)

        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self(x)
        loss = F.mse_loss(y_pred, y)

        # Some logs
        self.log('val/loss', loss, prog_bar=True)

        return loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_pred = self(x)
        loss = F.mse_loss(y_pred, y)

        # Some logs
        self.log('test/loss', loss, prog_bar=True)

        return loss



class SDKN(Network):
    '''
    Implementation of SDKN with initialization, forward and configure_optimizers method.
    '''

    def __init__(self, centers, learning_rate=1e-3):
        super().__init__()

        # Some settings
        self.register_buffer("centers", centers)
        self.M = self.centers.shape[0]
        self.learning_rate = learning_rate

        # Define linear maps - hardcoded
        self.fc1 = nn.Linear(5, 32, bias=False)
        self.fc2 = nn.Linear(32, 64, bias=False)
        self.fc3 = nn.Linear(64, 48, bias=False)
        self.fc4 = nn.Linear(48, 24, bias=False)
        self.fc5 = nn.Linear(24, 2, bias=False)

        torch.nn.init.kaiming_normal_(self.fc1.weight)
        torch.nn.init.kaiming_normal_(self.fc2.weight)
        torch.nn.init.kaiming_normal_(self.fc3.weight)
        torch.nn.init.kaiming_normal_(self.fc4.weight)
        torch.nn.init.kaiming_normal_(self.fc5.weight)

        # Define activation function mappings
        self.activ1 = ActivFunc(32, self.M, kernel=kernels.Gaussian())
        self.activ2 = ActivFunc(64, self.M, kernel=kernels.Gaussian())
        self.activ3 = ActivFunc(48, self.M, kernel=kernels.Gaussian())
        self.activ4 = ActivFunc(24, self.M, kernel=kernels.Gaussian())



    def forward(self, x):
        centers = self.centers

        # First fully connect + activation function
        x = self.fc1(x)
        centers = self.fc1(centers)
        x, centers = self.activ1(x, centers)

        # Second fully connect + activation function
        x = self.fc2(x)
        centers = self.fc2(centers)
        x, centers = self.activ2(x, centers)

        # Third fully connect + activation function
        x = self.fc3(x)
        centers = self.fc3(centers)
        x, centers = self.activ3(x, centers)

        # Fourth fully connect + activation function
        x = self.fc4(x)
        centers = self.fc4(centers)
        x, centers = self.activ4(x, centers)

        # Fifth (final) fully connect
        x = self.fc5(x)

        return x


    def configure_optimizers(self):         # Adam + LR scheduler
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer, step_size=settings.decayEpochs_sdkn, gamma=settings.decayRate)

        return [optimizer], [scheduler]



class NN(Network):
    '''
    Implementation of NN with initialization, forward and configure_optimizers method.
    '''

    def __init__(self, learning_rate=1e-3):
        super().__init__()

        # Some settings
        self.learning_rate = learning_rate

        # Define linear maps - hardcoded
        self.fc1 = nn.Linear(5, 32, bias=False)
        self.fc2 = nn.Linear(32, 64, bias=False)
        self.fc3 = nn.Linear(64, 48, bias=False)
        self.fc4 = nn.Linear(48, 24, bias=False)
        self.fc5 = nn.Linear(24, 2, bias=False)

        torch.nn.init.kaiming_normal_(self.fc1.weight)
        torch.nn.init.kaiming_normal_(self.fc2.weight)
        torch.nn.init.kaiming_normal_(self.fc3.weight)
        torch.nn.init.kaiming_normal_(self.fc4.weight)
        torch.nn.init.kaiming_normal_(self.fc5.weight)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = F.relu(self.fc3(x))
        x = F.relu(self.fc4(x))
        x = self.fc5(x)

        return x


    def configure_optimizers(self):         # Adam + LR scheduler
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer, step_size=settings.decayEpochs_nn, gamma=settings.decayRate)

        return [optimizer], [scheduler]